import json
import pickle
from pathlib import Path

from metadocker.common.errors import MetadockerError
from metadocker.common.misc import flatten

NoneType=None.__class__

class Empty: pass

EmptyValue = Empty()

def NotEmpty(x):
    if not x and x is not None:
        raise Var.ValidationException(None, f"doit être non vide")
    return x

def get_types(types):
    if not isinstance(types, (tuple, list, set)):
        types = [types]
    return tuple([x.__class__ if not isinstance(x, type) else x for x in types])

class EmptyObject:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

class Var:
    _expected_types_ = []
    def _validate(self, value):
        return value
    class VarException(MetadockerError):
        def get_error(self):
            raise NotImplementedError()

    class RequiredException(VarException):
        def __init__(self, name):
            super().__init__(name)
            self.name = name

        def get_error(self):
            return f"La variable {self.name} est requise"

    class ByPassException(MetadockerError):
        pass

    class ValidationException(VarException):
        def __init__(self, name, err):
            super().__init__(name, err)
            self.name = name
            self.err = err

        def get_error(self):
            return f"La variable {self.name} est invalide : {self.err}"

    class TypeException(VarException):
        def __init__(self, name, types, found):
            super().__init__(name, types)
            self.name = name

            if isinstance(types, type):
                types = [types]
            types = [x.__name__ for x in types]
            self.types = tuple(types)
            self.found = found

        def get_error(self):
            return f"La variable {self.name} doit être de type {'('+', '.join(self.types)+')'}. Type trouvé {self.found.__class__.__name__}"

    def __init__(self, name, default_value=EmptyValue, required=False, validate=None,
                 expected_types=None, help=None, is_meta=False, null=False, ask=False, **kwargs):
        self.name = name
        self.default_value = default_value
        self.required = required
        if validate:
            self._validate = validate
        self.expected_types = get_types(flatten(self._expected_types_ or [],  expected_types or []))
        self.help = help
        self.null = null or self.default_value is None
        self.is_meta = is_meta
        self.ask = ask
        self.kwargs = kwargs


    def on_error(self, obj, err):
        raise err

    @property
    def has_default(self):
        return self.default_value is not EmptyValue

    def run_validation(self, obj, is_parent = True):
        if not is_parent:
            obj = EmptyObject(**{self.name : obj})
        if not hasattr(obj, self.name):
            if self.default_value is EmptyValue:
                if self.required:
                    raise self.RequiredException(self.name)
                else:
                    raise self.ByPassException()
            return self.default_value

        data = getattr(obj, self.name)

        if self.expected_types:
            if data is None and self.null:
                pass
            elif not isinstance(data, self.expected_types):
                raise self.TypeException(self.name, self.expected_types, data)

            data = self._validate(data)

        return data

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"


    def run_to_python(self, value):
        if self.null and value == "None":
            return None
        return self.to_python(value)

    def to_python(self, value):
        if not self.expected_types:
            return value
        for t in self.expected_types:
            try:
                v = t(value)
                return self.run_validation(v, is_parent=False)
            except Exception as err:
                continue

        if self.null and not value:
            return self.default_value
        raise self.ValidationException(self.name, f"Impossible de parser '{value}'")

    def ask_value(self, **kwargs):
        has_value = "value" in kwargs
        value = kwargs.get("value")
        while True:
            print(self.help)
            prompt = [self.name]
            has_default = self.default_value is not EmptyValue

            if has_default:
                prompt.append(f" [{self.default_value}]")

            if not has_value:
                value = input("".join(prompt)+": ")
            else:
                print("".join(prompt) + ":", value)
                has_value = False


            if not has_default and not value:
                print(f"Merci de donner une réponse")
                continue

            if value=="":
                return self.default_value

            try:
                return self.run_to_python(value)
            except MetadockerError as err:
                if isinstance(err, self.ValidationException):
                    err.name = self.name
                print(err.get_error())


class VarStr(Var):
    _expected_types_ = (str, )

class VarPath(Var):
    _expected_types_ = (Path, str,)

class VarInt(Var):
    _expected_types_ = (int,)

class VarNumber(Var):
    _expected_types_ = (float, int,)

class VarDict(Var):
    _expected_types_ = (dict,)
    def  to_python(self, value):
        if isinstance(value, (bytes, str)):
            try:
                return self.run_validation(json.loads(value))
            except:
                pass
            if isinstance(value, bytes):
                try:
                    return self.run_validation(pickle.loads(value))
                except:
                    pass
        return super().to_python(value)

class VarList(Var):
    _expected_types_ = (list,)


