import json
import os
import pickle
from pathlib import Path

from metadocker.common.errors import SimpleError, MultipleError, MetadockerError
from metadocker.common.var import VarPath, VarList, EmptyObject, Var, VarDict, NotEmpty, VarInt


class DockerManagerBase:
    pass
class VarDockerManager(Var):
    _expected_types_ = (DockerManagerBase,)

    def on_error(self, obj, err):
        from metadocker.db.manager import DockerManager
        return DockerManager()


class _Config:
    _engine = pickle
    _INSTANCE = None

    APP__DIR = "app.dir"
    APP__DOCKER_CACHE = "app.docker.cache"
    APP__NEXT_PORT = "app.next_port"

    _vars_ = [
        VarPath(APP__DIR, required=True, validate=NotEmpty, help="Le dossier qui contient les données (configuration de metadocker)"),
        VarDockerManager(APP__DOCKER_CACHE, default_value={}, hidden=True),
        VarInt(APP__NEXT_PORT, default_value=8201, required=True)
    ]


    def __init__(self):
        if _Config._INSTANCE is not None:
            raise ValueError(f"Impossible de lancer pluseir base de données")
        self._var_dict_ = {x.name: x for x in self._vars_}
        self.filename = self._get_path() / f"metadocker.{self._engine.__name__}"
        self.data = None

    def load(self):
        if self.data is not None: return
        try:
            self.data = self._engine.loads(self.filename.read_bytes())
        except Exception as err:
            print(f"Impossible de trouver la configuration dans {self.filename}. Création")
            self.save()

        return self


    def _get_path(self):
        return Path(os.environ.get("METADOCKER_CONF_DIR",
                                   Path(os.environ["HOME"]) / ".config" / "metadocker" )).resolve()


    def save(self):
        self.filename.parent.mkdir(exist_ok=True, parents=True)
        self.filename.write_bytes(self._engine.dumps(self.data))

    def __getitem__(self, item):
        return self.data[item]
    def __setitem__(self, item, value):
        if item in self._var_dict_:
            var = self._var_dict_[item]
            value = var.to_python(value)
        self.data[item] = value
        self.save()

    def items(self):
        return self.data.items()

    def __contains__(self, item):
        return item in self.data

    def values(self):
        return self.data.values()

    def get(self, k, *args):
        return self.data.get(k, *args)

    def __iter__(self):
        return iter(self.data)

    def _prop(self, name, example=None):
        if name not in self.data:
            value = [f"Option {name} non définie."]
            if example: value.append(f"metadocker config \#a{name}={example}\"")
            raise SimpleError("\n".join(value))
        return self.data[name]

    @property
    def app_dir(self) -> Path: return self._prop(self.APP__DIR, "/dossier/de/done")
    @property
    def app_docker_cache(self) -> DockerManagerBase: return self._prop(self.APP__DOCKER_CACHE)


    def validate(self, raise_exception=True):
        errors = []
        for var in self._vars_:
            obj = EmptyObject()
            if var.name in self.data:
                setattr(obj, var.name, self.data[var.name])
            try:
                self.data[var.name] = var.run_validation(obj)
            except MetadockerError as err:
                try:
                    self.data[var.name] = var.on_error(obj, err)
                    self.save()
                except MetadockerError as err:
                    errors.append(err)

                if var.kwargs.get("on_error"):
                    var.kwargs["on_error"]()

        if errors and raise_exception:
            raise MultipleError(errors)
        self.save()
        return errors


settings = _Config()